/*
 *
 * Copyright 2013-2018 Radian Memory Systems, Inc.  All Rights Reserved.
 *
 * This file is provided under non-disclosure agreement.
 * See the file COPYING for details.
 *
 */
#ifndef DIALOG_DEF_H_
#define DIALOG_DEF_H_

/*
 * Extention of Log Page command Vendor specified log identifiers
 */
#define GLP_ID_DIALOG		            (0xD1)

/*
 * Definitions for the Dialog module which provides on-board collection of statistics across 
 * the life of the board.
*/

/* 
 * The values to be measured
 */
typedef enum
{
	DM_MEASUREMENTS_E_FIRST = 0,
	DM_PRINCETON_TEMP = DM_MEASUREMENTS_E_FIRST, /* 1. Temperature of the Princeton Controller */
	DM_CAP12_TEMP,				/* 2. Temperature of Capacitors 1 and 2 */
	DM_CAP34_TEMP,				/* 3. Temperature of Capacitors 3 and 4 */
	DM_AMBIENT_TEMP,			/* 4. Rear vent ambient temperature */
	DM_BOARD_TEMP,				/* 5. Temperature of RMS 200 board */
	DM_CAP1_VOLTAGE,			/* 6. Voltage of Capacitor 1 relative to ground */
	DM_CAP2_VOLTAGE,			/* 7. Voltage of Capacitor 2 relative to Capacitor 1 */
	DM_CAP3_VOLTAGE,			/* 8. Voltage of Capacitor 3 relative to Capacitor 2 */
	DM_PACK_VOLTAGE,			/* 9. Voltage of Capacitor 4 relative to ground (i.e., pack voltage) */
	DM_PACK_VOLT_F2F_END,		/* 10. Capacitor pack voltage at end of F2F */
	DM_V3_V4_DIFF,				/* 11. Current derived from v3-v4 */
	DM_DERIVED_POWER,			/* 12. Derived energy */
	DM_DERIVED_CAPACITANCE,		/* 13. Derived capacitance of the pack */
	DM_DERIVED_RESISTANCE,  	/* 14. Derived series resistance of capacitor pack */
	DM_F2F_TIME,	   		 	/* 15. Time to run F2F */
	DM_RESTORE_TIME,			/* 16. Time to run Restore */
	DM_CAP_CHARGE_TIME,			/* 17. Time to charge capacitors */
	DM_FLASH_PAGE_READ_TIME,	/* 18. Flash page read time */
	DM_FLASH_PAGE_WRITE_TIME,	/* 19. Flash page write time */
	DM_FLASH_BLOCK_ERASE_TIME,	/* 20. Flash block erase time */
	DM_CORRECTABLE_PAGE_BITS,	/* 21. Correctable bits in error on reading a page */
	DM_CORRECTABLE_BCH_BITS,	/* 22. Correctable bits in error on reading the worst BCH region of a page */
	DM_FAN_INLET_AMBIENT_TEMP,	/* 23. Fan inlet ambient temperature */
	DM_STD_MEASUREMENTS_LAST,	/* 24. Standard measurements in all products, also marks the start of extension. */
	DM_CAP4_VOLTAGE = DM_STD_MEASUREMENTS_LAST, /* 24. Voltage of Capacitor 4 relative to Capacitor 3 */
	DM_FAN_VOLTAGE,								/* 25. Voltage of the Fan sensor */
	/* .. */
	DM_ALL_MEASUREMENTS_E_LAST	/* Keep this to be the last */
} DialogMeasurementsEnum;

#define	DM_DERIVED_ENERGY	DM_DERIVED_POWER

#define	DIALOG_NUM_MEASUREMENTS	DM_STD_MEASUREMENTS_LAST
#define	DIALOG_NUM_MFLAGS	((DIALOG_NUM_MEASUREMENTS-1)/32+1)
#define	DIALOG_NUM_VERSIONS	5
#define	DIALOG_DESC_LEN_MAX	22

/* 
 * Events
 */
typedef enum
{
	DE_F2F,						/* 1. F2F */
	DE_PROPER_SHUTDOWN,			/* 2. Clean shutdown without F2F */
	DE_ERASE_ERROR,				/* 3. Error on erase */
	DE_WRITE_ERROR,				/* 4. Error on write */
	DE_UNCORRECTABLE_ERROR,		/* 5. Uncorrectable error on read */
	DE_RESTORE_INCOMPLETE,		/* 6. Restored incomplete flash image */
	DE_PRODUCTION,				/* 7. Production/Repair */
	/*
	 * a. Any capacitor has a voltage over 2.85V for more than 10 seconds
	 * b. Or the pack voltage exceeds the charge voltage by more than 0.4V.
	 * c. Data value recorded should be split into bit fields to allow recording both the voltage and which
	 *    capacitor (one through four, and use zero to indicate the full pack).
	 * d. An Asynchronous Event Notification indicating a permanent error should be sent to the host. 
	 */ 
	DE_EXCESSIVE_CAP_VLOTAGE,	/* 8. Excessive capacitor voltage */
	/* 
	 * a. During charging, current is outside the range [-10mA, +2.5A]
	 * b. Or during discharging, current is outside the range [-2.5A, +10mA]
	 * c. Or during normal operation, current is outside the range [-10mA, +10mA]
	 * d. Charging is defined as the time from power-up to whichever comes first, 5 minutes or current is in the
	 *    range [-5mA, +5mA] for 10 seconds.
	 * e. Data value recorded should be the current.
	 * f. An Asynchronous Event Notification indicating a permanent error should be sent to the host.
	 */
	DE_EXCESSIVE_CURRENT		/* 9. Excessive current */
} DialogEventsEnum;

#define	DIALOG_NUM_EVENTS		9

/*
 * Thresholds
*/
typedef enum
{
	DE_LEVEL_0,					/* Level 0: Error.  Exceptionally low value, perhaps suggesting permanent damage to the board. */
	DE_LEVEL_1,					/* Level 1: Error.  Unacceptably low value. */
	DE_LEVEL_2,					/* Level 2: Lower than typical value, but acceptable. */
	DE_LEVEL_3,					/* Level 3: Normal  value. */
	DE_LEVEL_4,					/* Level 4: Normal  value. */
	DE_LEVEL_5,					/* Level 5: Normal  value. */ 
	DE_LEVEL_6,					/* Level 6: Normal  value. */
	DE_LEVEL_7,					/* Level 7: Error.  Higher than typical value, but acceptable. */
	DE_LEVEL_8,					/* Level 8: Error.  Unacceptably high value. */
	DE_LEVEL_9					/* Level 9: Error.  Exceptionally high value, perhaps suggesting permanent damage to the board. */
} DialogThresholdsEnum;

#define	DIALOG_NUM_LEVELS		10
#define	DIALOG_NUM_THRESHOLDS	(DIALOG_NUM_LEVELS-1)

#define	DIALOG_WINDOW_SIZE	8

typedef struct _DialogSpan {
	u32		start;
	u32		duration : 24;
	u32		worst : 4;
} DialogSpan;

typedef struct _DialogMeasure {
	int		recent;
	unsigned int	bin[DIALOG_NUM_LEVELS];
	DialogSpan	span[DIALOG_WINDOW_SIZE];
} DialogMeasurement;

typedef struct _DialogLog {
	unsigned int	time;
	int		value;
} DialogLog;

typedef	struct _DialogEvent {
	unsigned int		count;
	DialogLog		log[DIALOG_WINDOW_SIZE];
} DialogEvent;

typedef union _DialogPageExtensions {
	u32			padding[79]; 		/* Align to 54 * 64 */
	struct dialog_extensions {
		u32 nDwords;		/* Number of u32's remaining in this struct */
		u8 firmwareRev[8];
		u8 hardwareRev[8];
		u16 capLife[2];
		u64 avg_temp_cap[2];	/* for internal use only */
		u64 avg_temp_proc[2];	/* for internal use only */
		u32 nFanLogs;
		u32 cap4recent;		/* In case of a 5 Capacitors pack */
		u32 chargeMultiplier;  	/* Either 3.87 or 5.32 multiplied by 1000 */
		struct dialog_config_desc {
			u8 format_id;
			u8 version;
			u8 description[DIALOG_DESC_LEN_MAX];
		} desc;
		/* used to indicate to host whether the measurement is valid or not */
		u32 mflags[DIALOG_NUM_MFLAGS];
	} extensions;
} __attribute__((__packed__)) DialogPadding;
typedef char PADDING_SIZE_BUG[sizeof(DialogPadding) > (79*4) ? -1:1];

typedef struct _DialogPage {
	int				dialogVersion;
	u32				snapshotTime;

	DialogMeasurement measurements[DIALOG_NUM_MEASUREMENTS];

	DialogEvent		events[DIALOG_NUM_EVENTS];
	u32				nextSnapshotTime;

	u32				refVcc; 			/* Refernce voltage of the sensor for temperature calculations */
	u32				initialCapacitance; /* Initial measured capacitance */
	u32				initialEsr; 		/* Initial measured Equivalent Series Resistance */
	u32				reserved;
	u32				errorConditionExists; /* bit flied indicating varoius error conditions */
	u32				adjustedCapacitance; /* Adjustement to the measured capacitance */
	u32				lowMarginAdjustment; /* An adjustement in case the cap voltage after F2F it too low */

        /* Additions to the dialog struct should be placed in the 'padding' area
         * to maintin compatibility with earlier versions of dialog.
         */
	DialogPadding		dpad;

	struct 
	{
		u32 criticalWarn : 8;
		u32 temperature : 16;
		u32 availableSpare : 8;
		u8 availableSpareThrd;
		u8 pctUsed;
		u8 rsvd31_06[26];
		u64 dataUnitReadLow;
		u64 dataUnitReadHigh;
		u64 dataUnitWriteLow;
		u64 dataUnitWriteHigh;
		u64 hostReadCmdLow;
		u64 hostReadCmdHigh;
		u64 hostWriteCmdLow;
		u64 hostWriteCmdHigh;
		u64 ctrlBusyTimeLow;
		u64 ctrlBusyTimeHigh;
		u64 pwrCyclLow;
		u64 pwrCyclHigh;
		u64 pwrOnHourLow;
		u64 pwrOnHourHigh;
		u64 unsaveShutdownLow;
		u64 unsaveShutdownHigh;
		u64 mediaErrLow;
		u64 mediaErrHigh;
		u64 numOfErrLogLow;
		u64 numOfErrLogHigh;
		u8 rsvd511_192[320];
	} smartData;  /* 512 bytes @ 54 * 64 = 3456 */
} DialogPage;

typedef enum {
    SET_TIME_IN_SECONDS = 0x0,
    CLEAR_DIALOG_PAGE,
    TAKE_A_SNAP_SHOT,
    ENTER_PRODUCTION_CODE,
    DO_CAP_ESR_MEASUREMENT_CODE,
	SET_ERROR_CONDITION,
	DIALOG_CFG_RELOAD /* firmware internal use, from admin manager to dialog manager */
} DialogUserCodesEnum;

struct dialog_err_level {
	u32	exceed:4;
	u32 below:4;
	u32 reserved:24;
};

typedef struct dialog_err_level dialog_err_level_t;

/*
 * Temperature is in C integers 
 * Voltage is in mili Volts 
 * Current is in mili Ampers 
 * Resistance is in milil Ohms 
 * Capacitance is in mili Farads 
*/
/* thermistor Steinhart-Hart coefficients */
typedef struct steinhart_param {
	float coef_a;
	float coef_b;
	float coef_c;
} steinhart_t;

/* CAP/ESR range limit and conversion parameters */
typedef struct {
	double voltage_conversion_k;
	double current_conversion_k;
	float pack_to_charge_multiplier;
	u16 validCapacitanceLow;
	u16 validCapacitanceHigh;
	u16 defaultAdjustedCapacitance;
	u16 validEsrLow;
	u16 validEsrHigh;
	u16 validEsrTyp;
}  __attribute__((__packed__)) range_lc_t;

/* Joule curve parameters */
typedef struct {
	/*
	 * Coefficients for quadratic form for charge voltage setting as a function of
	 * processor temperature
	 */
	float t2jCoefA;
	float t2jCoefB;
	float t2jCoefC;
	/*
	 * Limits on the above equation
	 */
	u16	minJoules;
	u16	maxJoules;

	/* temperature trigger for override of quadratic in tempToJoules() */
	u16 minJoulesTemp;
	u16 maxJoulesTemp;
} joule_curve_t;

/* capacitance calibration parameters */
typedef struct {
	float cap_temp_norm;
	float cap_temp_slope;
	float cap_voltage_norm;
	float cap_voltage_slope;
} cap_calibration_t;

/* capacitance measurement control */
typedef struct {
	float capMeasureBump;
	u16 caps_min_wait_charge_interval;
	u16 caps_wait_charge_interval;
	u8 integratePercentage;
	u8 chargevTrimPercentage;
} cap_measure_control_t;

/* charge voltage control parameters */
typedef struct {
	/*
	 * Typical expected time to run F2F, which relates to the power and thus the
	 * Joules needed
	 */
	float chargeInSeconds;
	/*
	 * Low voltage at which we can't keep running F2F, and its square for
	 * convenience
	 */
	float cchg_vmin;
	/*
	 * Highest capacitor pack voltage we are willing to set.
	 * This is the manufacturer's specified maximum voltage for the capacitor pack.
	 */
	float cchg_vmax;
	float default_charge_voltage;
	float minimum_charge_voltage;
	float cchg_operatingVmin;
	float lowMarginTrigger;
	float charge_hysteresis;
	float charge_margin;
} voltage_control_t;

/* Capacitor lifetime parameters */
typedef struct {
	/*
	 * Minimum capacitance at which we can confidently run F2F even with rising
	 * temperatures
	 */
	float 	minCapacitance;
	/*
	 * Pack capacitance after initial break-in (close to nominal capacitance)
	 */
	float 	capElbow; /* CL_ELBOW */

	float cl_vmax;
	float cl_vt;
	float cl_voff;
	float cl_tshift;
	float cl_tdep;
	float cl_rfactor;
	float cl_lifemax;
} capacitor_life_t;

typedef struct board_specific_param {
	steinhart_t		shh;
	range_lc_t		rlc;
	joule_curve_t	jcurv;
	cap_calibration_t		ccal;
	cap_measure_control_t	cmc;
	voltage_control_t		vc;
	capacitor_life_t		clife;
	u16				flag1;
	u16				flag2;
} board_specific_t;

struct dialog_config {
	u16 verify_crc;
	u16 pad;
	/* histogram thesholds */
	dialog_err_level_t	error_levels[DIALOG_NUM_MEASUREMENTS];
	u16 thresholds[DIALOG_NUM_MEASUREMENTS][DIALOG_NUM_THRESHOLDS];
	/* misc temperature thresholds */
	u16 capacitor_over_heat;			  /* 60.0 degrees */   
	u16 peltier_turnon;                   /* 45.0 degrees */   
	u16 peltier_turnoff;                  /* 35.0 degrees */   
	u16 peltier_turnoff_cpu;              /* 90.0 degrees */   
	u16 ambient_warning;                  /* 55.0 degrees */   
	u16 capacitor_temp;                   /* 60.0 degrees */   
	u16 cpu_overtemp_alert;               /* 100.0 degrees */  
	u16 cpu_overtemp_shutdown;            /* 110.0 degrees */  
	/* other misc thresholds */             
	u16 charge_voltage_max;               /* 11.00 V */      
	u16 F2F_voltage_min;                  /* 4.50 V */       
	u16 capacitor_single_voltage;         /* 2.85 V */       
	u16 capacitor_single_voltage_time;    /* 10 seconds */   
	u16 capacitor_pack_overvoltage;       /* 0.40 V */       
	u16 current_max_charging;             /* 2.500 A */      
	u16 current_max_discharging;          /* 2.500 A */      
	u16 current_max_normal;               /* 0.010 A */      
	u16 F2F_max_energy;                   /* 220 J */        
	u16 post_F2F_minimum_energy;           /* 95 J */         
	u16 minimum_farads;                   /* 4.25 F */       
	u16 maximum_internal_resistance;      /* 0.500 Ohms */   
	/* Capacitance Depletion Rate Parameters */
	u16 reserved1[5];
	u16 cap_pack_rated_farads;                        /* 6.25 F */     
	u16 reserved2[4];
	/* On board hardware settings */
	u16 digital_potentiometer_value;                  /* LSB wr0 MSB wr1  */

	/* board and component specific stuff */
	board_specific_t bs;

	/* dialog configuration description */
	u8 format_id;
	u8 version;
	u8 description[DIALOG_DESC_LEN_MAX]; /* null-terminated string */
};

typedef struct dialog_config dialog_config_t;

/* compile time message if the struct is getting too big to fit into the EEPROM header area */
typedef char BUILD_STRUCT_SIZE_BUG[sizeof(dialog_config_t) <= (1024 - 256) ? 1:-1];

#endif /* DIALOG_DEF_H_ */

